from pydfs_lineup_optimizer.sites.draftkings.classic.importer import DraftKingsCSVImporter
from pydfs_lineup_optimizer.sites.draftkings.classic.settings import *
from pydfs_lineup_optimizer.sites.draftkings.captain_mode.importer import DraftKingsCaptainModeCSVImporter
from pydfs_lineup_optimizer.sites.draftkings.captain_mode.settings import *
from pydfs_lineup_optimizer.sites.draftkings.tiers.settings import *

__all__ = [
    'DraftKingsCSVImporter', 'DraftKingsBasketballSettings', 'DraftKingsFootballSettings',
    'DraftKingsHockeySettings', 'DraftKingsBaseballSettings', 'DraftKingsGolfSettings',
    'DraftKingsSoccerSettings', 'DraftKingsCanadianFootballSettings', 'DraftKingsLOLSettings',
    'DraftKingsCaptainModeCSVImporter', 'DraftKingsCaptainModeFootballSettings',
    'DraftKingsCaptainModeBasketballSettings', 'DraftKingsTiersBasketballSettings',
]



from collections import namedtuple
from datetime import datetime
from pytz import timezone
from typing import List, Optional
from pydfs_lineup_optimizer.utils import process_percents
from pydfs_lineup_optimizer.tz import get_timezone


GameInfo = namedtuple('GameInfo', ['home_team', 'away_team', 'starts_at', 'game_started'])


class Player:
    def __init__(self,
                 player_id: str,
                 first_name: str,
                 last_name: str,
                 positions: List[str],
                 team: str,
                 salary: float,
                 fppg: float,
                 is_injured: bool = False,
                 max_exposure: Optional[float] = None,
                 min_exposure: Optional[float] = None,
                 projected_ownership: Optional[float] = None,
                 game_info: Optional[GameInfo] = None,
                 roster_order: Optional[int] = None,
                 min_deviation: Optional[float] = None,
                 max_deviation: Optional[float] = None,
                 is_confirmed_starter: Optional[bool] = None,
                 fppg_floor: Optional[float] = None,
                 fppg_ceil: Optional[float] = None,
                 original_positions: Optional[List[str]] = None,
                 ):
        self.id = player_id
        self.first_name = first_name
        self.last_name = last_name
        self.positions = positions
        self.team = team
        self.salary = salary
        self.fppg = fppg
        self.is_injured = is_injured
        self.game_info = game_info
        self.roster_order = roster_order
        self._min_exposure = None  # type: Optional[float]
        self._max_exposure = None  # type: Optional[float]
        self._min_deviation = None  # type: Optional[float]
        self._max_deviation = None  # type: Optional[float]
        self._projected_ownership = None  # type: Optional[float]
        self.min_exposure = min_exposure
        self.max_exposure = max_exposure
        self.min_deviation = min_deviation
        self.max_deviation = max_deviation
        self.projected_ownership = projected_ownership
        self.is_confirmed_starter = is_confirmed_starter
        self.fppg_floor = fppg_floor
        self.fppg_ceil = fppg_ceil
        self._original_positions = original_positions

    def __repr__(self):
        return '%s %s (%s)' % (self.full_name, '/'.join(self.positions), self.team)

    def __hash__(self):
        return hash(self.id)

    @property
    def max_exposure(self) -> Optional[float]:
        return self._max_exposure

    @max_exposure.setter
    def max_exposure(self, max_exposure: Optional[float]):
        self._max_exposure = process_percents(max_exposure)

    @property
    def min_exposure(self) -> Optional[float]:
        return self._min_exposure

    @min_exposure.setter
    def min_exposure(self, min_exposure: Optional[float]):
        self._min_exposure = process_percents(min_exposure)

    @property
    def min_deviation(self) -> Optional[float]:
        return self._min_deviation

    @min_deviation.setter
    def min_deviation(self, min_deviation: Optional[float]):
        self._min_deviation = process_percents(min_deviation)

    @property
    def max_deviation(self) -> Optional[float]:
        return self._max_deviation

    @max_deviation.setter
    def max_deviation(self, max_deviation: Optional[float]):
        self._max_deviation = process_percents(max_deviation)

    @property
    def projected_ownership(self) -> Optional[float]:
        return self._projected_ownership

    @projected_ownership.setter
    def projected_ownership(self, projected_ownership: Optional[float]):
        self._projected_ownership = process_percents(projected_ownership)

    @property
    def full_name(self) -> str:
        return '{} {}'.format(self.first_name, self.last_name)

    @property
    def efficiency(self) -> float:
        return round(self.fppg / self.salary, 6)

    @property
    def is_game_started(self) -> bool:
        if self.game_info:
            if self.game_info.game_started:
                return True
            starts_at = self.game_info.starts_at
            time_now = datetime.now().replace(tzinfo=timezone(get_timezone()))
            if starts_at and time_now > starts_at:
                return True
        return False

    @property
    def original_positions(self) -> List[str]:
        return self._original_positions or self.positions


class LineupPlayer:
    def __init__(self, player: Player, lineup_position: str, used_fppg: Optional[float] = None):
        self._player = player
        self.lineup_position = lineup_position
        self.used_fppg = used_fppg

    def __getattr__(self, attr_name):
        return getattr(self._player, attr_name)

    def __eq__(self, other):
        if isinstance(other, Player):
            return self._player == other
        elif isinstance(other, LineupPlayer):
            return self._player == other._player
        return NotImplemented

    def __hash__(self):
        return hash(self._player)

    def __repr__(self):
        return repr(self._player)



from pydfs_lineup_optimizer.solvers.base import Solver
from pydfs_lineup_optimizer.solvers.pulp_solver import PuLPSolver
from pydfs_lineup_optimizer.solvers.constants import SolverSign
from pydfs_lineup_optimizer.solvers.exceptions import SolverException, SolverInfeasibleSolutionException


__all__ = ['Solver', 'PuLPSolver', 'SolverSign', 'SolverException', 'SolverInfeasibleSolutionException']



from typing import TYPE_CHECKING, DefaultDict, Dict
from collections import Counter, defaultdict, OrderedDict
from itertools import chain
from pydfs_lineup_optimizer.player import LineupPlayer


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.lineup_optimizer import LineupOptimizer


class Statistic:
    def __init__(self, optimizer: 'LineupOptimizer'):
        self.optimizer = optimizer
        if self.optimizer.last_context is not None:
            self.lineups = self.optimizer.last_context.lineups
        else:
            self.lineups = []

    def get_top_teams(self) -> Dict[str, int]:
        teams_dict = defaultdict(int)  # type: DefaultDict[str, int]
        for lineup in self.lineups:
            lineup_teams = {p.team for p in lineup}
            for team in lineup_teams:
                teams_dict[team] += 1
        return OrderedDict(sorted(teams_dict.items(), key=lambda t: -t[1]))

    def get_top_players(self) -> Dict[str, Dict[LineupPlayer, int]]:
        players = defaultdict(lambda: defaultdict(int))  # type: DefaultDict[str, DefaultDict[LineupPlayer, int]]
        for player in chain.from_iterable(self.lineups):
            players[player.full_name][player] += 1
        return OrderedDict(sorted(players.items(), key=lambda t: -sum(t[1].values())))  # type: ignore

    def print_report(self) -> None:
        top_teams = self.get_top_teams()
        if len(top_teams) > 1:
            print('Top Teams')
            for team, appearance in top_teams.items():
                print('  %s - %d' % (team, appearance))
            print('Used %d/%d\n' % (len(top_teams), len(self.optimizer.available_teams)))
        print('Top Players')
        players_per_team = defaultdict(dict)  # type: DefaultDict[str, Dict[str, int]]
        top_players = self.get_top_players()
        total_players = len({player.full_name for player in self.optimizer.players})
        replicated_players = Counter([player.full_name for player in self.optimizer.players])
        for player_name, players in top_players.items():
            total = sum(players.values())
            players_per_team[list(players.keys())[0].team][player_name] = total
            by_positions = ''
            if replicated_players[player_name] > 1:
                by_positions = ','.join('%s-%d' % ('/'.join(player.positions), appearance)
                                        for player, appearance in players.items())
                by_positions = '(%s)' % by_positions
            print('  %s - %d %s' % (player_name, total, by_positions))
        print('Used %d/%d\n' % (len(top_players), total_players))
        if len(top_teams) > 1:
            print('Top Players Per Team')
            for team, team_players in sorted(players_per_team.items(), key=lambda t: -sum(t[1].values())):
                total_appearances = sum(team_players.values())
                team_players_str = ','.join('%s(%d)' % values for values in
                                            sorted(players_per_team[team].items(), key=lambda t: -t[1]))
                print('  %s(%d) - %s' % (team, total_appearances, team_players_str))



from typing import List, Dict, Any, Optional
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.lineup import Lineup


class CSVImporter:
    def __init__(self, filename: str):
        self.filename = filename

    def import_players(self) -> List[Player]:  # pragma: no cover
        raise NotImplementedError

    def import_lineups(self, players: List[Player]) -> List[Lineup]:  # pragma: no cover
        pass

    @staticmethod
    def _parse_exposure(exposure: Optional[str]) -> Optional[float]:
        exposure = (exposure or '').replace('%', '')
        return float(exposure) if exposure else None

    @classmethod
    def get_player_extra(cls, row: Dict[str, str]) -> Dict[str, Any]:
        roster_order = row.get('Roster Order')
        fppg_floor = row.get('Projection Floor')
        fppg_ceil = row.get('Projection Ceil')
        return {
            'max_exposure': cls._parse_exposure(row.get('Max Exposure')),
            'min_exposure': cls._parse_exposure(row.get('Min Exposure')),
            'roster_order': int(roster_order) if roster_order else None,
            'projected_ownership': cls._parse_exposure(row.get('Projected Ownership')),
            'min_deviation': cls._parse_exposure(row.get('Min Deviation')),
            'max_deviation': cls._parse_exposure(row.get('Max Deviation')),
            'is_confirmed_starter': bool(row.get('Confirmed Starter')),
            'fppg_floor': float(fppg_floor) if fppg_floor else None,
            'fppg_ceil': float(fppg_ceil) if fppg_ceil else None,
        }



from typing import TypeVar, Any, List, Iterable, Optional


Self = TypeVar('Self', bound='Solver')


class Solver:  # pragma: no cover
    def setup_solver(self) -> None:
        raise NotImplementedError

    def set_objective(self, variables: Iterable[Any], coefficients: Iterable[float]):
        raise NotImplementedError

    def add_variable(self, name: str, min_value: Optional[int] = None, max_value: Optional[int] = None) -> Any:
        raise NotImplementedError

    def add_constraint(self, variables: Iterable[Any], coefficients: Optional[Iterable[float]], sign: str, rhs: float,
                       name: Optional[str] = None):
        raise NotImplementedError

    def solve(self) -> List[Any]:
        raise NotImplementedError

    def copy(self) -> Self:
        raise NotImplementedError



from pydfs_lineup_optimizer.sites.fanduel.classic.importer import *
from pydfs_lineup_optimizer.sites.fanduel.classic.settings import *
from pydfs_lineup_optimizer.sites.fanduel.single_game.importer import *
from pydfs_lineup_optimizer.sites.fanduel.single_game.settings import *


__all__ = [
    'FanDuelCSVImporter', 'FanDuelBasketballSettings', 'FanDuelFootballSettings', 'FanDuelHockeySettings',
    'FanDuelBaseballSettings', 'FanDuelWnbaSettings',
    'FanDuelMVPCSVImporter', 'FanDuelSingleGameFootballSettings', 'FanDuelSingleGameBasketballSettings',
]



import csv
from typing import Iterable, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.lineup import Lineup
    from pydfs_lineup_optimizer.player import LineupPlayer


class LineupExporter:
    def __init__(self, lineups: Iterable['Lineup']):
        self.lineups = lineups

    @staticmethod
    def render_player(player: 'LineupPlayer') -> str:
        result = player.full_name  # type: str
        if player.id:
            result += '(%s)' % player.id
        return result

    def export(self, filename: str, render_func: Callable[['LineupPlayer'], str] = None):
        raise NotImplementedError


class CSVLineupExporter(LineupExporter):
    def export(self, filename, render_func=None):
        with open(filename, 'w') as csvfile:
            lineup_writer = csv.writer(csvfile, delimiter=',')
            for index, lineup in enumerate(self.lineups):
                if index == 0:
                    header = [player.lineup_position for player in lineup.lineup]
                    header.extend(('Budget', 'FPPG'))
                    lineup_writer.writerow(header)
                row = [(render_func or self.render_player)(player) for player in lineup.lineup]
                row.append(str(lineup.salary_costs))
                row.append(str(lineup.fantasy_points_projection))
                lineup_writer.writerow(row)


class FantasyDraftCSVLineupExporter(LineupExporter):
    def export(self, filename, render_func=None):
        if not self.lineups:
            return
        total_players = 0
        with open(filename, 'r') as csvfile:
            lines = list(csv.reader(csvfile))
            for i, lineup in enumerate(self.lineups, start=1):
                if i >= len(lines):
                    lines.append([])
                players_list = [(render_func or self.render_player)(player) for player in lineup.lineup]
                if not total_players:
                    total_players = len(players_list)
                lines[i] = players_list + lines[i][total_players:]
            for line_order in range(i, len(lines) - 1):
                lines[line_order] = [''] * total_players + lines[line_order][total_players:]
        with open(filename, 'w') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(lines)


class YahooCSVLineupExporter(CSVLineupExporter):
    @staticmethod
    def render_player(player: 'LineupPlayer') -> str:
        return str(player.id)



from typing import Optional, Type, List, Sequence, TYPE_CHECKING
from pydfs_lineup_optimizer.lineup_printer import LineupPrinter, BaseLineupPrinter
from pydfs_lineup_optimizer.lineup_exporter import LineupExporter, CSVLineupExporter


if TYPE_CHECKING:  # pragma: no cover
    from pydfs_lineup_optimizer.rules import OptimizerRule
    from pydfs_lineup_optimizer.lineup_importer import CSVImporter


class LineupPosition:
    def __init__(self, name: str, positions: Sequence[str]):
        self.name = name
        self.positions = positions


class BaseSettings:
    site = None  # type: str
    sport = None  # type: str
    budget = 0  # type: Optional[float]
    positions = []  # type: List[LineupPosition]
    max_from_one_team = None  # type: Optional[int]
    min_teams = None  # type: Optional[int]
    min_games = None  # type: Optional[int]
    total_teams_exclude_positions = []  # type: List[str]
    lineup_printer = LineupPrinter  # type: Type[BaseLineupPrinter]
    extra_rules = []  # type: List[Type['OptimizerRule']]
    csv_importer = None  # type: Type['CSVImporter']
    csv_exporter = CSVLineupExporter  # type: Type[LineupExporter]

    @classmethod
    def get_total_players(cls) -> int:
        return len(cls.positions)
